@@ -1,17 +1,23 @@
 // SPDX-License-Identifier: GPL-2.0-only
 
 #include <assert.h>
+#include <wlr/types/wlr_cursor.h>
 #include <wlr/types/wlr_fractional_scale_v1.h>
+#include <wlr/types/wlr_scene.h>
+#include <wlr/types/wlr_xdg_activation_v1.h>
+#include <wlr/types/wlr_xdg_shell.h>
 #include <wlr/types/wlr_xdg_toplevel_icon_v1.h>
 #include "buffer.h"
 #include "common/array.h"
 #include "common/macros.h"
 #include "common/mem.h"
+#include "config/rcxml.h"
 #include "decorations.h"
-#include "foreign-toplevel.h"
+#include "foreign-toplevel/foreign.h"
 #include "labwc.h"
 #include "menu/menu.h"
 #include "node.h"
+#include "output.h"
 #include "snap-constraints.h"
 #include "view.h"
 #include "view-impl-common.h"
@@ -63,7 +69,7 @@ xdg_toplevel_view_get_size_hints(struct view *view)
 
 static bool
 xdg_toplevel_view_contains_window_type(struct view *view,
-		enum window_type window_type)
+		enum lab_window_type window_type)
 {
 	assert(view);
 
@@ -75,9 +81,9 @@ xdg_toplevel_view_contains_window_type(struct view *view,
 		|| toplevel->parent;
 
 	switch (window_type) {
-	case NET_WM_WINDOW_TYPE_NORMAL:
+	case LAB_WINDOW_TYPE_NORMAL:
 		return !is_dialog;
-	case NET_WM_WINDOW_TYPE_DIALOG:
+	case LAB_WINDOW_TYPE_DIALOG:
 		return is_dialog;
 	default:
 		return false;
@@ -285,6 +291,14 @@ handle_configure_timeout(void *data)
 	view->pending_configure_serial = 0;
 	view->pending_configure_timeout = NULL;
 
+	/*
+	 * No need to do anything else if the view is just being slow to
+	 * map - the map handler will take care of the positioning.
+	 */
+	if (!view->mapped) {
+		return 0; /* ignored per wl_event_loop docs */
+	}
+
 	bool empty_pending = wlr_box_empty(&view->pending);
 	if (empty_pending || view->pending.x != view->current.x
 			|| view->pending.y != view->current.y) {
@@ -311,6 +325,13 @@ handle_configure_timeout(void *data)
 			wlr_log(WLR_INFO, "using fallback position");
 			view->pending.x = VIEW_FALLBACK_X;
 			view->pending.y = VIEW_FALLBACK_Y;
+			/* At least try to keep it on the same output */
+			if (output_is_usable(view->output)) {
+				struct wlr_box box =
+					output_usable_area_in_layout_coords(view->output);
+				view->pending.x += box.x;
+				view->pending.y += box.y;
+			}
 		}
 		view->current.x = view->pending.x;
 		view->current.y = view->pending.y;
@@ -380,7 +401,7 @@ handle_request_move(struct wl_listener *listener, void *data)
 	 */
 	struct view *view = wl_container_of(listener, view, request_move);
 	if (view == view->server->seat.pressed.view) {
-		interactive_begin(view, LAB_INPUT_STATE_MOVE, 0);
+		interactive_begin(view, LAB_INPUT_STATE_MOVE, LAB_EDGE_NONE);
 	}
 }
 
@@ -575,8 +596,7 @@ xdg_toplevel_view_append_children(struct view *self, struct wl_array *children)
 	struct wlr_xdg_toplevel *toplevel = xdg_toplevel_from_view(self);
 	struct view *view;
 
-	wl_list_for_each_reverse(view, &self->server->views, link)
-	{
+	wl_list_for_each_reverse(view, &self->server->views, link) {
 		if (view == self) {
 			continue;
 		}
@@ -635,7 +655,7 @@ xdg_toplevel_view_notify_tiled(struct view *view)
 		return;
 	}
 
-	enum wlr_edges edge = WLR_EDGE_NONE;
+	enum lab_edge edge = LAB_EDGE_NONE;
 
 	bool want_edge = rc.snap_tiling_events_mode & LAB_TILING_EVENTS_EDGE;
 	bool want_region = rc.snap_tiling_events_mode & LAB_TILING_EVENTS_REGION;
@@ -646,27 +666,33 @@ xdg_toplevel_view_notify_tiled(struct view *view)
 	 */
 	if (want_edge) {
 		switch (view->tiled) {
-		case VIEW_EDGE_LEFT:
-			edge = WLR_EDGE_LEFT | WLR_EDGE_TOP | WLR_EDGE_BOTTOM;
+		case LAB_EDGE_LEFT:
+			edge = LAB_EDGES_EXCEPT_RIGHT;
 			break;
-		case VIEW_EDGE_RIGHT:
-			edge = WLR_EDGE_RIGHT | WLR_EDGE_TOP | WLR_EDGE_BOTTOM;
+		case LAB_EDGE_RIGHT:
+			edge = LAB_EDGES_EXCEPT_LEFT;
 			break;
-		case VIEW_EDGE_UP:
-			edge = WLR_EDGE_TOP | WLR_EDGE_LEFT | WLR_EDGE_RIGHT;
+		case LAB_EDGE_TOP:
+			edge = LAB_EDGES_EXCEPT_BOTTOM;
 			break;
-		case VIEW_EDGE_DOWN:
-			edge = WLR_EDGE_BOTTOM | WLR_EDGE_LEFT | WLR_EDGE_RIGHT;
+		case LAB_EDGE_BOTTOM:
+			edge = LAB_EDGES_EXCEPT_TOP;
 			break;
+		case LAB_EDGES_TOP_LEFT:
+		case LAB_EDGES_TOP_RIGHT:
+		case LAB_EDGES_BOTTOM_LEFT:
+		case LAB_EDGES_BOTTOM_RIGHT:
+			edge = view->tiled;
+			break;
+		/* TODO: LAB_EDGE_CENTER? */
 		default:
-			edge = WLR_EDGE_NONE;
+			edge = LAB_EDGE_NONE;
 		}
 	}
 
 	if (want_region && view->tiled_region) {
 		/* Region-snapped views are considered tiled on all edges */
-		edge = WLR_EDGE_LEFT | WLR_EDGE_RIGHT |
-			WLR_EDGE_TOP | WLR_EDGE_BOTTOM;
+		edge = LAB_EDGES_ALL;
 	}
 
 	uint32_t serial =
@@ -999,7 +1025,7 @@ handle_new_xdg_toplevel(struct wl_listener *listener, void *data)
 	}
 	view->content_tree = tree;
 	node_descriptor_create(&view->scene_tree->node,
-		LAB_NODE_DESC_VIEW, view);
+		LAB_NODE_VIEW, view, /*data*/ NULL);
 
 	/*
 	 * The xdg_toplevel_decoration and kde_server_decoration protocols
@@ -1055,9 +1081,6 @@ static void
 handle_xdg_toplevel_icon_set_icon(struct wl_listener *listener, void *data)
 {
 	struct wlr_xdg_toplevel_icon_manager_v1_set_icon_event *event = data;
-
-	struct server *server =
-		wl_container_of(listener, server, xdg_toplevel_icon_set_icon);
 	struct wlr_xdg_surface *xdg_surface = event->toplevel->base;
 	struct view *view = xdg_surface->data;
 	assert(view);
